#include "application_manager.h"
#include "application_list_model.h"
#include "application.h"
#include "desktopfilereader.h"
#include "logging.h"

#include "shellserverconfiguration.h"
#include "sessionlistener.h"

#include "mir/shell/session_manager.h"

#include <QDebug>
#include <QProcess>

ApplicationManager::ApplicationManager(ShellServerConfiguration* server, QObject *parent)
:   QObject(parent)
,   m_mainStageApplications(new ApplicationListModel())
,   m_mainStageFocusedApplication(NULL)
,   m_mirServer(server)
{
    DLOG("ApplicationManager::ApplicationManager (this=%p)", this);
    qRegisterMetaType<ApplicationManager*>("ApplicationManager*"); //need if queueing signals

    std::shared_ptr<SessionListener> sessionListener
            = std::dynamic_pointer_cast<SessionListener>(m_mirServer->the_shell_session_listener());

    QObject::connect(sessionListener.get(), &SessionListener::sessionStarting,
                     this, &ApplicationManager::sessionStarting);
    QObject::connect(sessionListener.get(), &SessionListener::sessionStopping,
                     this, &ApplicationManager::sessionStopping);
    QObject::connect(sessionListener.get(), &SessionListener::sessionFocused,
                     this, &ApplicationManager::sessionFocused);
    QObject::connect(sessionListener.get(), &SessionListener::sessionUnfocused,
                     this, &ApplicationManager::sessionUnfocused);
}

ApplicationManager::~ApplicationManager()
{
    DLOG("ApplicationManager::~ApplicationManager");
    delete m_mainStageApplications;
}

ApplicationListModel* ApplicationManager::mainStageApplications() const
{
    DLOG("ApplicationManager::mainStageApplications (this=%p)", this);
    return m_mainStageApplications;
}

Application* ApplicationManager::mainStageFocusedApplication() const
{
    DLOG("ApplicationManager::mainStageFocusedApplication (this=%p)", this);
    return m_mainStageFocusedApplication;
}

void ApplicationManager::focusApplication(Application* application)
{
    DLOG("ApplicationManager::focusApplication (this=%p, application=%p)", this, application);

    m_mirServer->the_session_manager()->set_focus_to(application->session());
}

void ApplicationManager::focusFavoriteApplication(
        ApplicationManager::FavoriteApplication application)
{
    DLOG("ApplicationManager::focusFavoriteApplication (this=%p, application=%d)",
         this, static_cast<int>(application));
    // TODO(greyback) - upstart should handle favourite applications
}

void ApplicationManager::unfocusCurrentApplication(ApplicationManager::StageHint stageHint)
{
    DLOG("ApplicationManager::unfocusCurrentApplication (this=%p, stageHint=%d)", this,
         static_cast<int>(stageHint));

    m_mirServer->the_session_manager()->set_focus_to(NULL); //FIXME(greyback)
}

Application* ApplicationManager::startProcess(QString desktopFile, ApplicationManager::ExecFlags flags,
                                           QStringList arguments)
{
    DLOG("ApplicationManager::startProcess (this=%p, flags=%d)", this, (int) flags);
    // Load desktop file.
    DesktopFileReader* desktopData = new DesktopFileReader(desktopFile);
    if (!desktopData->loaded()) {
        delete desktopData;
        return NULL;
    }

    arguments.prepend("APP_ID=" + desktopData->appId());
    arguments.prepend("application");

    // Start process.
    //FIXME(greyback) use upstart via DBUS
    QProcess process;
    process.start("start", arguments);
    //FIXME(greyback) this is blocking, instead listen for finished() signal
    process.waitForFinished(1000);

    DLOG_IF(process.error(), "process 'start application' failed to start with error: %s", process.errorString().toLatin1().data());
    if (!process.error()) {
        // fetch pid
        QString output = process.readAllStandardOutput();
        int lastSpace = output.lastIndexOf(' ');
        int pid = output.remove(0, lastSpace).toInt();

        DLOG("process started with pid %d, adding '%s' to application lists", pid, desktopData->name().toLatin1().data());
        Application* application = new Application(
                    desktopData, pid,
                    (flags.testFlag(ApplicationManager::ForceMainStage)
                        || desktopData->stageHint() != "SideStage")
                        ? Application::MainStage : Application::SideStage,
                    Application::Starting //FIXME(greyback): assuming running immediately
                    );

        m_mainStageApplications->add(application);
        return application;
    } else {
        return NULL;
    }
}

void ApplicationManager::stopProcess(Application* application)
{
    DLOG("ApplicationManager::stopProcess (this=%p, application=%p)", this, application);

    if (application != NULL) {
        m_mainStageApplications->remove(application);

        // Start process.
        bool result;
        result = QProcess::startDetached("stop", QStringList() << "application"
                                         << ("APP_ID=" + application->appId()));
        DLOG_IF(result == false, "process 'stop application' failed to start");

        delete application;
    }
}

/************************************* Mir-side methods *************************************/

QString ApplicationManager::authorizedSessionName(const qint64 pid)
{
    DLOG("ApplicationManager::authorizedSessionName (this=%p, pid=%lld)", this, pid);
    Application* application = m_mainStageApplications->getApplicationWithPid(pid);
    if (application) {
        QString name = application->appId() + QString::number(qrand());
        application->setSessionName(name);
        return name;
    } else {
        return QString(); // i.e. do not authorize //TODO(greyback) dislike
    }
}

void ApplicationManager::sessionStarting(std::shared_ptr<mir::shell::ApplicationSession> const& session)
{
    DLOG("ApplicationManager::sessionStarting (this=%p, application=%s)", this, session->name().c_str());
    Application* application = m_mainStageApplications->getApplicationWithSession(
                                    QString::fromStdString(session->name()));
    if (application) {
        application->setSession(session);
        application->setState(Application::Running);
    } else {
        DLOG("ApplicationManager::sessionStarting - unauthorized application!!");
    }
}

void ApplicationManager::sessionStopping(std::shared_ptr<mir::shell::ApplicationSession> const& session)
{
    DLOG("ApplicationManager::sessionStopping (this=%p, application=%s)", this, session->name().c_str());

    // in case application closed not by hand of shell, check again here:
    Application* application = m_mainStageApplications->getApplicationWithSession(
                                    QString::fromStdString(session->name()));
    if (application) {
        if (application == m_mainStageFocusedApplication) {
            // TODO(greyback) either focus next app, or unfocus everything
        }
        m_mainStageApplications->remove(application);
    }
}

void ApplicationManager::sessionFocused(std::shared_ptr<mir::shell::ApplicationSession> const& session)
{
    DLOG("ApplicationManager::sessionFocused (this=%p, application=%s)", this, session->name().c_str());
    Application* application = m_mainStageApplications->getApplicationWithSession(
                                    QString::fromStdString(session->name()));
    if (application && application != m_mainStageFocusedApplication) {
        m_mainStageFocusedApplication = application;
        m_mainStageFocusedApplication->setFocus(true);
        Q_EMIT mainStageFocusedApplicationChanged();
    }
}

void ApplicationManager::sessionUnfocused()
{
    DLOG("ApplicationManager::sessionUnfocused (this=%p)", this);
    if (NULL != m_mainStageFocusedApplication) {
        m_mainStageFocusedApplication->setFocus(false);
        m_mainStageFocusedApplication = NULL;
        Q_EMIT mainStageFocusedApplicationChanged();
    }
}

