#ifndef APPLICATIONMANAGER_H
#define APPLICATIONMANAGER_H

// std
#include <memory>

// Qt
#include <QObject>
#include <QStringList>

#include "application.h"
#include "application_list_model.h"

class ShellServerConfiguration;

namespace mir { namespace shell { class ApplicationSession; }}

class ApplicationManager : public QObject
{
    Q_OBJECT
    Q_ENUMS(Role)
    Q_ENUMS(StageHint)
    Q_ENUMS(FormFactorHint)
    Q_ENUMS(FavoriteApplication)
    Q_FLAGS(ExecFlags)

    Q_PROPERTY(ApplicationListModel* mainStageApplications READ mainStageApplications
               NOTIFY mainStageApplicationsChanged)
    Q_PROPERTY(Application* mainStageFocusedApplication READ mainStageFocusedApplication
               NOTIFY mainStageFocusedApplicationChanged)

public:
    explicit ApplicationManager(ShellServerConfiguration* server, QObject *parent = 0);
    ~ApplicationManager();

    // Mapping enums to Ubuntu Platform API enums.
    enum Role {
      Dash, Default, Indicators, Notifications, Greeter,
      Launcher, OnScreenKeyboard, ShutdownDialog
    };
    enum StageHint {
      MainStage, IntegrationStage, ShareStage, ContentPickingStage,
      SideStage, ConfigurationStage
    };
    enum FormFactorHint {
      DesktopFormFactor, PhoneFormFactor, TabletFormFactor
    };
    enum FavoriteApplication {
      CameraApplication, GalleryApplication, BrowserApplication, ShareApplication
    };
    enum Flag {
      NoFlag = 0x0,
      ForceMainStage = 0x1,
    };
    Q_DECLARE_FLAGS(ExecFlags, Flag)

    ApplicationListModel* mainStageApplications() const;
    Application* mainStageFocusedApplication() const;

    Q_INVOKABLE void focusApplication(Application *application);
    Q_INVOKABLE void focusFavoriteApplication(FavoriteApplication application);
    Q_INVOKABLE void unfocusCurrentApplication(StageHint stageHint);
    Q_INVOKABLE Application* startProcess(QString desktopFile, ExecFlags flags = NoFlag, QStringList arguments = QStringList());
    Q_INVOKABLE void stopProcess(Application* application);

    QString authorizedSessionName(const qint64 pid);

Q_SIGNALS:
    void mainStageApplicationsChanged();
    void mainStageFocusedApplicationChanged();
    void focusRequested(FavoriteApplication favoriteApplication);

public Q_SLOTS:
    void sessionStarting(std::shared_ptr<mir::shell::ApplicationSession> const& session);
    void sessionStopping(std::shared_ptr<mir::shell::ApplicationSession> const& session);
    void sessionFocused(std::shared_ptr<mir::shell::ApplicationSession> const& session);
    void sessionUnfocused();

private:
    ApplicationListModel* m_mainStageApplications;
    Application* m_mainStageFocusedApplication; // remove as Mir has API for this
    ShellServerConfiguration* m_mirServer;
};

Q_DECLARE_OPERATORS_FOR_FLAGS(ApplicationManager::ExecFlags)

Q_DECLARE_METATYPE(ApplicationManager*)

#endif // APPLICATIONMANAGER_H
